#!/bin/bash
# Generated by runner ({{ runnerVersion }}) at {{ timestamp }}
set -e

# Files created at run-time should be group-writable
umask 0002

function prepend-with-date {
  while read line
  do
    echo $(date +%Y-%m-%dT%H:%M:%S%z) $line
  done
}

{% set instance_data = jsonMap(container.instanceData()) %}
{% set cluster_data = jsonMap(container.clusterData()) %}
{% set esMajorVersion = version.major() %}
{% set esMinorVersion = version.minor() %}
{% set esPatchVersion = version.patch() %}
{% set supportsPurge = esMajorVersion >= 6 or (esMajorVersion == 5 and esMinorVersion >= 5) %}
# In 7.15.1+ we use JVM version 17 which supports in-JVM compression of heap dumps
{% set supportsInJvmHeapDumpCompression = esMajorVersion > 7 or (esMajorVersion == 7 and esMinorVersion > 15) or (esMajorVersion == 7 and esMinorVersion == 15 and esPatchVersion >= 1) %}

# Since 7.7.0 we've switched to use upstream images which have different file location
if [ -d /usr/share/elasticsearch ]; then
    ELASTICSEARCH_HOME=/usr/share/elasticsearch
else
    ELASTICSEARCH_HOME=/elasticsearch
fi

{% if esMajorVersion == 1 %}
INSTALL='/elasticsearch/bin/plugin -install'
{% elif esMajorVersion == 2 %}
INSTALL='/elasticsearch/bin/plugin install'
{% else %}
INSTALL=${ELASTICSEARCH_HOME}'/bin/elasticsearch-plugin install --silent --batch'
{% endif %}

if [[ $(id -u) -eq 0 ]]; then
    echo "ERROR: elasticsearch.sh running as uid 0 (root), exiting."
    exit 1
fi

export ES_PATH_CONF="${ES_PATH_CONF:-/app/config}"
export CONF_DIR=$ES_PATH_CONF # bwc for 1.x and bin/elasticsearch-plugin on 5.x
export LOG_DIR="/app/logs"
export DATA_DIR="/app/data"
export USER_BUNDLES_DIR="/app/found_user_bundles"
export USER_BUNDLES_SCRIPTS_DIR="$USER_BUNDLES_DIR/scripts"
export USER_BUNDLES_DICTS_DIR="$USER_BUNDLES_DIR/dictionaries"

# ES_TMPDIR is introduced in ES 6.2
# Please refer to https://www.elastic.co/guide/en/elasticsearch/reference/current/important-settings.html#es-tmpdir
# The exact folder name (`tmp`) should be the one that is specified in no.found.runner.allocation.GenericDockerContainer.tempPath
export ES_TMPDIR="/app/tmp"
mkdir -p ${ES_TMPDIR}
rm -rf ${ES_TMPDIR}/*

{% if esMajorVersion >= 7 %}
XPACK_ROOT="/app/config"
{% elif esMajorVersion == 6 and esMinorVersion >= 3 %}
XPACK_ROOT="/app/config"
{% elif esMajorVersion >= 5 %}
XPACK_ROOT="/app/config/x-pack"
{% endif %}

cd /app

mkdir -p logs
BOOT_LOG_FILE=logs/boot.log
# save stdout and stderr for later
exec 3>&1 4>&2
exec >  >(prepend-with-date | tee -a $BOOT_LOG_FILE)
exec 2> >(prepend-with-date | tee -a $BOOT_LOG_FILE >&2)


# The lock logic is removed from elasticsearch.sh in `cloud` repo.
# Still we have to keep it here - stack packs can be used with an older ECE
# versions that still rely on the lock file to get created.

# File to indicate all setup has been completed by this script so that the allocator
# knows when to move forward with other configuration such as managing the keystore.
# By using a distinct signal, this ensures that changes made within this script are not
# run concurrently with changes made by the allocator.
{% set script_ready_lock = scriptReadyLock|default("_script_lock_not_specified") %}

# Remove the lock prior to creating the boot log so that the allocator can
# wait for boot logs to indicate the script has started running and then subsequently
# wait for this script to be ready as keystore calls are made from both the allocator
# and this script!
rm -f {{ script_ready_lock }}

echo "Booting at $(date)"

# If the number of container processors is determined in the allocator as per:
# https://github.com/elastic/cloud/issues/44434 we just need to set the number of active processors
{% set supports_processors = containerProcessors|default(-1) %}
{% if supports_processors > 0 and ((esMajorVersion == 7 and esMinorVersion >= 8) or esMajorVersion > 7) %}
export ES_JAVA_OPTS="-XX:ActiveProcessorCount={{ containerProcessors }}"
{% else %}
# Number of GC threads should depend on the size of the node.
# Tiered compilation does not help Elasticsearch much, and reserves 48M memory for additional code caches.
# Disabling it reduces the code cache, however, which is problematic: https://github.com/elastic/cloud/issues/4603
export ES_JAVA_OPTS="-XX:ParallelGCThreads={{ container.parallelGCThreads() }} -XX:ConcGCThreads={{ container.concGCThreads() }}"
{% endif %}

{% if enforceV4|default(false) %}
echo "Enforcing SigV4"
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dcom.amazonaws.services.s3.enforceV4"
{% endif %}

# Before 7.4 we need this temporary patch below because of upstream Java setup changes (ref: https://github.com/elastic/cloud/issues/27895)
# Since 7.4 our images are using JDK distribution bundled within Elasticsearch and having JAVA_HOME system-wide is no longer needed (ref: https://github.com/elastic/cloud/issues/33907)
{% if esMajorVersion == 7 and esMinorVersion < 4 %}
if [ -z "$JAVA_HOME" ]; then
   export JAVA_HOME=$(dirname $(dirname $(readlink -f $(which java))))
fi
{% endif %}

{% if esMajorVersion >= 6 %}
    {% if cluster_data.aws_resources.s3.access_key_id and cluster_data.aws_resources.s3.secret_access_key %}
        {% set snapshot_client = cluster_data.aws_resources.s3.client_name|default("default") %}
        rm -f /app/config/elasticsearch.keystore
        ${ELASTICSEARCH_HOME}/bin/elasticsearch-keystore create --silent
        {% set s3_access_key_id = cluster_data.aws_resources.s3.access_key_id | replace("'", "") %}
        echo "{{ s3_access_key_id }}" | ${ELASTICSEARCH_HOME}/bin/elasticsearch-keystore add --stdin s3.client.{{ snapshot_client }}.access_key
        {% set s3_secret_access_key = cluster_data.aws_resources.s3.secret_access_key | replace("'", "") %}
        echo "{{ s3_secret_access_key }}" | ${ELASTICSEARCH_HOME}/bin/elasticsearch-keystore add --stdin s3.client.{{ snapshot_client }}.secret_key
    {% endif %}

    {% if cluster_data.gcp_resources.gcs %}
        {% set snapshot_client = cluster_data.gcp_resources.gcs.client_name|default("default") %}
        rm -f /app/config/elasticsearch.keystore
        ${ELASTICSEARCH_HOME}/bin/elasticsearch-keystore create --silent
        export SERVICE_ACCOUNT_FILE="/app/config/service_account.json"
        echo '{{ serviceAccount }}' > $SERVICE_ACCOUNT_FILE
        ${ELASTICSEARCH_HOME}/bin/elasticsearch-keystore add-file gcs.client.{{ snapshot_client }}.credentials_file $SERVICE_ACCOUNT_FILE
        rm $SERVICE_ACCOUNT_FILE
    {% endif %}

    cp ${ELASTICSEARCH_HOME}/config/jvm.options /app/config/jvm.options
    {% if esMajorVersion < 8 %}
    echo "" >> /app/config/jvm.options
    echo "# Allow aws access key and secret key to exist in the repository settings for backwards compatibility with old ECE versions supported by 7.x series" >> /app/config/jvm.options
    echo "-Des.allow_insecure_settings=true" >> /app/config/jvm.options
    {% endif %}

    {% if (esMajorVersion == 7 and esMinorVersion >= 2) or esMajorVersion >= 8 %}
        # On 7.2 we switched to jdk11 which changed the logging and by default logs to gc.log and caused
        # that those logs would be pushed to the logging and metrics cluster which is undesired
        sed -i -e 's!file=logs/gc.log!file=/app/logs/gc.output!g' /app/config/jvm.options
        # additionally, lets only keep 2 gc log files around, rather than 32
        sed -i -E -e 's!tags:filecount=32!tags:filecount=2!g' /app/config/jvm.options
        # and drop the size from 64MB to 8MB
        sed -i -E -e 's!filesize=64m!filesize=8m!g' /app/config/jvm.options
    {% else %}
        {% if (esMajorVersion == 6 and esMinorVersion >= 2) or esMajorVersion >= 7 %}
            # 6.2 does gc logging by default, but to /elasticsearch/logs, we use /app/logs instead
            sed -i -e 's!logs/gc.log!/app/logs/gc.log!g' /app/config/jvm.options
            # additionally, lets only keep 2 gc log files around, rather than 32
            sed -i -E -e 's!NumberOfGCLogFiles=32!NumberOfGCLogFiles=2!g' /app/config/jvm.options
            # and drop the size from 64MB to 8MB
            sed -i -E -e 's!FileSize=64m!FileSize=8m!g' /app/config/jvm.options
        {% endif %}
    {% endif %}
{% endif %}

{% if esMajorVersion <= 2 %}
    export ES_MAX_MEM={{ container.heapSize() }}M
    export ES_MIN_MEM={{ container.heapSize() }}M
    export ES_HEAP_SIZE={{ container.heapSize() }}M
{% else %}
    {% if automaticHeapSizingEnabled  %}
        export ES_JAVA_OPTS="$ES_JAVA_OPTS -Des.total_memory_bytes={{ container.availableMemoryInBytes() }}"
        {%- if hasHeapOverride|default(false) %}
        export ES_JAVA_OPTS="$ES_JAVA_OPTS -Xms{{ container.heapSize() }}M -Xmx{{ container.heapSize() }}M"
        {%- endif %}
    {% else %}
        export ES_JAVA_OPTS="$ES_JAVA_OPTS -Xms{{ container.heapSize() }}M -Xmx{{ container.heapSize() }}M"
    {% endif %}

    {% if container.heapSize() < 2048 %}
        # The heap isn't big, so switch to unpooled allocator in netty to save some heap
        export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dio.netty.allocator.type=unpooled"
    {% endif %}
{% endif %}

# Tuning for 5.x per https://github.com/elastic/cloud/issues/6153
# NIO poorly handles freeing direct buffers via the garbage collector.  Fixed in 5.5.1+
{% if esMajorVersion == 5 or esMajorVersion == 2 %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:+ExplicitGCInvokesConcurrent"
    {% if esMajorVersion == 2 or esMinorVersion <= 4 or (esMinorVersion == 5 and esPatchVersion == 0) %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:-DisableExplicitGC"
    {% endif %}
{% endif %}

# Newer JVMs that we bundle with ES 7.4+ include a UseContainerSupport option which is enabled by default.
# This causes Runtime.availableProcessors() to report unexpectedly low processor counts.
# Disable this setting and let it use all cpus on the host (up to quota) as with earlier JVMs.
# This option is no longer included in 7.12.1+ - see https://github.com/elastic/cloud/issues/80566
{% if (esMajorVersion == 7 and esMinorVersion >= 4 and (esMinorVersion < 12 or (esMinorVersion == 12 and esPatchVersion == 0))) %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:-UseContainerSupport"
{% endif %}

{% if (esMajorVersion == 7 and esMinorVersion >= 15) or esMajorVersion > 7 %}
#Cloud has some awareness attributes that are the same across zones and as such it gives better performance to
#ignore awareness attributes in searches so they don't disable Adaptive Replica Selection.
#Customers that prefer reduced interzone traffic over speed can disable ARS through cluster settings api.
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Des.search.ignore_awareness_attributes=true"
{% endif %}

{% if esMajorVersion < 6 %}
# In versions before 6.0.0, use quota-aware-fs if included in this image.
if [ -f /elasticsearch/lib/quota-aware-fs-*.jar ]; then
    echo "Enabling QuotaAwareFileSystemProvider"
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -Djava.nio.file.spi.DefaultFileSystemProvider=co.elastic.cloud.quotaawarefs.QuotaAwareFileSystemProvider"
fi
{% elif esMajorVersion == 7 and esMinorVersion >= 11 and esMinorVersion < 14 %}
# While not affected by the original bug, all versions that loaded quotaware fs as a plugin when released must still have it configured or the plugin blocks boot.
# Images after 6.0 that only loaded the jar doesn't need to keep loading it.
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Des.fs.quota.file=file:///app/config/fsquota.properties"
{% endif %}

# 5.0+ handles OOMs itself https://github.com/elastic/elasticsearch/pull/19272
{% if esMajorVersion <= 2 %}
# ES docker images built before 2016-02-26 may not have libjvmkill installed.
if [ -e /usr/lib/libjvmkill.so ]; then
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -agentpath:/usr/lib/libjvmkill.so"
else
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:OnOutOfMemoryError=/app/on-error-exitcode.sh"
fi
{% endif %}

{% if heapDumpOnOOM|default(false) %}
if [ "{{heapDumpEnabledUntil}}" -gt "$(date +%s)" ]; then
    echo "Enabling heap dump on oom"
    {% if supportsInJvmHeapDumpCompression %}
    # ES 7.15+ has JVM 17 which supports in-JVM heap dump compression, otherwise heap dump compression is provided by on-error-exitcode.sh
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:+HeapDumpOnOutOfMemoryError -XX:HeapDumpPath=/app/heap_dumps/compressed -XX:HeapDumpGzipLevel=6"
    {% else %}
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:+HeapDumpOnOutOfMemoryError -XX:HeapDumpPath=/app/heap_dumps"
    {% endif %}
else
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:-HeapDumpOnOutOfMemoryError"
fi
{% else %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:-HeapDumpOnOutOfMemoryError"
{% endif %}

# Less verbose curator
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dcurator-log-only-first-connection-issue-as-error-level=true"

{% if esMajorVersion < 5 %}
    export PLUGIN_TARGET_DIR="/data/plugins"
    export PLUGIN_SOURCE_DIR="/elasticsearch/plugins"
{% elif esMajorVersion >= 8 %}
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dio.netty.recycler.maxCapacityPerThread=0"
    export PLUGIN_TARGET_DIR="${ELASTICSEARCH_HOME}/plugins"
    export PLUGIN_SOURCE_DIR="/opt/plugins/archive"
{% else %}
    {#
        # In 5.x the ability to set plugin.path was removed
        # https://www.elastic.co/guide/en/elasticsearch/reference/5.0/breaking_50_plugins.html#_custom_plugins_path
    #}
    # ref: https://github.com/elastic/elasticsearch/pull/22452
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dio.netty.recycler.maxCapacityPerThread=0"
    export PLUGIN_TARGET_DIR="${ELASTICSEARCH_HOME}/plugins"
    export PLUGIN_SOURCE_DIR="${ELASTICSEARCH_HOME}/plugin-archives"
{% endif %}

{% if (esMajorVersion == 6 or (esMajorVersion == 7 and esMinorVersion < 3)) %}
    # ES 6 and after make use of the java security manager
    # this will create the gelf policy in order to allow gelf appender to work
    # Also when searching for the jar we remove the cli as it does not require
    # to have a custom security policy.
    es_jar=$(ls /elasticsearch/lib/elasticsearch*.jar | grep -v -E "(cli|core|launchers|secure-sm|x-content)")
    es_cli_jar=$(ls /elasticsearch/lib/elasticsearch-cli*.jar || true) # only 6.1.0 and up
    log4j_jar=$(ls /elasticsearch/lib/log4j-core*.jar)
    gelf_jar=$(ls /elasticsearch/lib/log4j2-gelf*.jar)
    echo "Generating gelf security policy file"
    rm -f /app/config/gelf.policy
    for jar in "$es_jar" "$es_cli_jar" "$log4j_jar" "$gelf_jar"; do
        if [ -n "$jar" ] && [ -e "$jar" ]; then
            # tee so the policy will appear in docker logs, too
            cat << POLICY | tee -a /app/config/gelf.policy
grant codeBase "file://${jar}" {
  permission java.net.SocketPermission "*", "connect,resolve";
};
POLICY
        fi
    done

    export ES_JAVA_OPTS="$ES_JAVA_OPTS -Djava.security.policy=file:///app/config/gelf.policy"
    # if we want to debug the security manager and log access and failures, also need to change the logging level
    # export ES_JAVA_OPTS="$ES_JAVA_OPTS -Djava.security.debug=\"access,failure\""
{% endif %}

{% if (esMajorVersion == 5 and ((esMinorVersion == 1 and esPatchVersion >= 2) or (esMinorVersion >= 2))) or (esMajorVersion >= 6) %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Des.cgroups.hierarchy.override=/"
{% endif %}

{% if (esMajorVersion == 6 and esMinorVersion >= 3) and (esMajorVersion == 6 and esMinorVersion < 7) %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Des.geoip.load_db_on_heap=true"
{% endif %}

{% set plugin_set_identifier = pluginSetIdentifier | replace('"', "") %}
export PLUGIN_SET_MARKER=${ELASTICSEARCH_HOME}"/found_installed_plugins-{{ plugin_set_identifier }}"
export LEGACY_PLUGIN_SET_MARKER="/data/found_installed_plugins-{{ plugin_set_identifier }}"
export USER_PLUGINS_MARKER_PATH=${ELASTICSEARCH_HOME}"/user_plugins_installed"
if [ ! -e $PLUGIN_SET_MARKER ] && [ ! -e $LEGACY_PLUGIN_SET_MARKER ]; then
    echo "First time booting with this image, verifying ephemeral directories are empty"

    # Assert the data folder doesn't contain any cruft
    # If we see any error we log but carry on
    if [ $(ls ${ELASTICSEARCH_HOME}/found_installed_plugins* 2> /dev/null | wc -l) -gt 0 ]; then
        echo "WARNING: Ephemeral data path contains marker file from different image, did Docker change behaviour? (continuing)"
    fi
    if [ $(ls $PLUGIN_TARGET_DIR/* 2> /dev/null | wc -l) -gt 0 ]; then
        echo "WARNING: Plugin install dir is not empty, but plugin marker file was missing (continuing)"
    fi

    rm -rf /app/plugins #This can line be removed when dropping backwards compatibility with ECE 1.0

    # If it is a symlink, make it a regular directory. Most 5.0 to 5.4 images have a symlink here, no longer needed.
    if [ -h ${ELASTICSEARCH_HOME}/plugins ]; then
        rm ${ELASTICSEARCH_HOME}/plugins
    fi
    mkdir -p $PLUGIN_TARGET_DIR

    export PURGE_CONFIG_POST_INSTALL='true'
fi

{% if esMajorVersion < 5 %}
    {% for plugin_name in instance_data.elasticsearch.plugins %}
        # if the plugin does not exist copy from bundle. Symlinking does not work since 2.x
        {% set clean_plugin_name = plugin_name | replace("'", "") %}
        if [ ! -e $PLUGIN_TARGET_DIR/{{ clean_plugin_name }} ]; then
            cp -r $PLUGIN_SOURCE_DIR/{{ clean_plugin_name }} $PLUGIN_TARGET_DIR/{{ clean_plugin_name }} || true
        fi
    {%  endfor %}
{% else %}
    # Deleting any marker file here as a workaround for supporting 5.x images
    # without ephemeral dir, otherwise listing plugins fail below.
    rm -f ${ELASTICSEARCH_HOME}/plugins/.found_installed_plugins*

    mkdir -p $USER_PLUGINS_MARKER_PATH

    # When listing the installed plugins, we remove any plugins that start with a tab character as these plugins are
    # part of a plugin bundle. When removing plugins, we would remove the bundle as Elasticsearch will then remove the
    # associated bundle plugins
    export ALREADY_INSTALLED=$(${ELASTICSEARCH_HOME}/bin/elasticsearch-plugin list | grep -v "^warning:" | sed '/^\t/d' | sort)

    export TO_BE_REMOVED=$(comm -23 <(echo "$ALREADY_INSTALLED") <(
        for PLUGIN in {{ allPluginsInstalled|replace("\n", "  ") | replace(";", "") }} ; do \
            MARKERFILE=$(find "$USER_PLUGINS_MARKER_PATH" -regex "$USER_PLUGINS_MARKER_PATH/$PLUGIN-[a-f0-9]+")
            if [ -n "$MARKERFILE" ] ; then
                cat "$MARKERFILE"
            else
                echo $PLUGIN
            fi
        done | sort
    ))
    {% set plugins_to_add_clean = instance_data.elasticsearch.plugins|join('\n') | replace("'", "") %}
    export TO_BE_ADDED=$(comm -13 <(echo "$ALREADY_INSTALLED") <(echo '{{ plugins_to_add_clean }}' | sort))

    for PLUGIN in $TO_BE_REMOVED; do \
        echo "Removing unselected plugin: $PLUGIN"
        {% if supportsPurge %}
            ${ELASTICSEARCH_HOME}/bin/elasticsearch-plugin remove --purge $PLUGIN
        {% else %}
            ${ELASTICSEARCH_HOME}/bin/elasticsearch-plugin remove $PLUGIN
            if [ -e /app/config/$PLUGIN ]; then
                echo "Purging config at: /app/config/$PLUGIN"
                rm -r "/app/config/$PLUGIN"
            fi
        {% endif %}
        # Wipe out any marker file that may have referenced this plugin
        grep -Rl "^$PLUGIN$" $USER_PLUGINS_MARKER_PATH | xargs rm -f
    done

    {% if esMajorVersion >= 8 or (esMajorVersion == 7 and esMinorVersion >= 6)  %}
        if [ ! -z "$TO_BE_ADDED" ] ; then
            PLUGIN_ARCHIVES_LIST=$(for PLUGIN in $TO_BE_ADDED; do echo file://$(readlink -f $PLUGIN_SOURCE_DIR/$PLUGIN-*.zip); done)
            echo "Adding following plugin files: $PLUGIN_ARCHIVES_LIST"
            $INSTALL $PLUGIN_ARCHIVES_LIST
            echo "Adding plugins done"
        fi
    {% else %}
        for PLUGIN in $TO_BE_ADDED; do \
            echo "Adding selected plugin: $PLUGIN"
            $INSTALL file://$(readlink -f $PLUGIN_SOURCE_DIR/$PLUGIN-*.zip)
        done
    {% endif %}
{% endif %}

# Precursor to installing bundles and plugins
mkdir -p "$USER_BUNDLES_DIR/.found-cached-downloads"

echo "Installing user plugins."
{% for user_plugin in instance_data.elasticsearch.user_plugins %}
    {% set name = user_plugin.name | replace("'", "") %}
    {% set url = user_plugin.url | replace("'", "") %}
    {% if url != user_plugin.url %}
      echo "Single quotes were replaced from the plugin's URL, this might result in an error when starting the container"
    {% endif %}
    # Installing user plugin: {{ name }}
    {% if esMajorVersion < 5 %} {# Version allows rename user plugins to configured name #}
        if [ ! -e $PLUGIN_TARGET_DIR/{{ name }} ] && [ ! -e $PLUGIN_TARGET_DIR/{{ name }}/.found_installed_plugin ]; then
            echo "Installing {{ name }}..."
            rm -rf $PLUGIN_TARGET_DIR/{{ name }} || true

            {% if esMajorVersion == 1 %}
                JAVA_OPTS=-Des.path.plugins=$PLUGIN_TARGET_DIR $INSTALL {{ name }} -url '{{ url }}' && touch $PLUGIN_TARGET_DIR/{{ name }}/.found_installed_plugin || true
            {% else %}
                wget --no-verbose -O $PLUGIN_TARGET_DIR/{{ name }}.zip '{{ url }}' \
                    && unzip -o $PLUGIN_TARGET_DIR/{{ name }}.zip -d $PLUGIN_TARGET_DIR/{{ name }} \
                    || false
                rm $PLUGIN_TARGET_DIR/{{ name }}.zip
            {% endif %}

            echo "Installed {{ name }}."
        else
            if [ -e $PLUGIN_TARGET_DIR/{{ name }}/.found_installed_plugin ]; then
                echo "Already installed {{ name }}."
            else
                echo "Not installing {{ name }} because it already exists as a built-in, enabled plugin"
            fi
        fi
    {% else %} {# Record real name in marker file #}
        MARKER_FILE="$USER_PLUGINS_MARKER_PATH/{{ name }}-$(echo '{{ url }}' | md5sum | awk '{ print $1 }' )"
        if [ ! -e "$MARKER_FILE" ]; then
            export OTHER_MARKER=$(find $USER_PLUGINS_MARKER_PATH -regex $USER_PLUGINS_MARKER_PATH/{{ name }}-[a-f0-9]+)
            if [ -n "$OTHER_MARKER" ]; then
                OTHER_MARKER_NAME=$(cat "$OTHER_MARKER")
                if [ -n "$OTHER_MARKER_NAME" ]; then
                    echo "Uninstalling previous version of user plugin. plugin=[{{ name }}], name=[$OTHER_MARKER_NAME]"
                    ${ELASTICSEARCH_HOME}/bin/elasticsearch-plugin remove $OTHER_MARKER_NAME
                else
                    echo "Failed to uninstall previous previous version of user plugin: Marker does not contain plugin name. plugin=[{{ name }}]"
                fi
                rm "$OTHER_MARKER"
            fi
            echo "Installing user plugin [{{ name }}]..."

            # Cache or download the plugin, then install:
            {% set clean_plugin_url = user_plugin['url'] | replace("'", "") %}
            {% if clean_plugin_url != user_plugin['url'] %}
              echo "A single quote was filtered from the URL of the plugin [{{ url }}]. This might result on a container not starting, but could be a signal of an attack"
            {% endif %}
            {% set clean_plugin_url_hash = clean_plugin_url | md5 %}
            PLUGIN_PREFIX="$USER_BUNDLES_DIR/.found-cached-downloads/plugin-{{ name }}"

            # (we'll use $USER_BUNDLES_DIR for the cache since we know that persists)
            if [ ! -e $PLUGIN_PREFIX-{{ clean_plugin_url_hash }}.zip ]; then
              echo "Downloading user plugin [{{ name }}]"

              if ! wget --no-use-server-timestamps --no-verbose -O $PLUGIN_PREFIX-{{ clean_plugin_url_hash }}.zip '{{ url }}'; then
                echo "Could not download user plugin [{{ name }}]. Will now exit and retry after restart."
                exit 1 {#
                  NOTE: It is better to abort and possibly bootloop on a broken plugin url, than to start ES and have
                  shards fail because of a missing plugin. The first is a real error outside of our responsibillity and
                  the latter could be us causing data loss because of a transient network issue.
                #}
              else
                echo "Clearing expired cached versions of user plugin [{{ name }}]"
                ls -t $PLUGIN_PREFIX-*.zip | tail -n +3 | xargs rm -f
              fi
            else
              echo "user plugin is cached, re-using that [{{ name }}]"
            fi

            # Now we have either downloaded or de-cached the plugin, install it:
            {% if esMajorVersion < 6 or (esMajorVersion == 6 and esMinorVersion < 3) %} {# Before 6.3, plugins must have an "elasticsearch" dir at the root #}
            declare PLUGIN_DESCRIPTOR_PATH=elasticsearch/plugin-descriptor.properties
            {% else %} {# Since 6.3, there is no "elasticsearch" root dir anymore #}
            # The file can be called plugin-descriptor.properties or stable-plugin-descriptor.properties
            declare PLUGIN_DESCRIPTOR_PATH="*plugin-descriptor.properties"
            {% endif %}
            declare REAL_NAME=$(
                unzip -p $PLUGIN_PREFIX-{{ clean_plugin_url_hash }}.zip "$PLUGIN_DESCRIPTOR_PATH" \
                | grep '^ *name *[=:]' | tr : = | cut -d'=' -f2- | awk '{$1=$1};1' | sed 's/\r$//'
            )
            if $INSTALL "file://$(readlink -f $PLUGIN_PREFIX-{{ clean_plugin_url_hash }}.zip)"; then
              echo "$REAL_NAME" > "$MARKER_FILE"
              echo "Installed user plugin [{{ name }}]."
            else
              echo "Plugin [{{ name }}] appears to be corrupt, deleting and exiting"
              rm -f "$PLUGIN_PREFIX-{{ clean_plugin_url_hash }}.zip"
              exit 1
            fi
        else
            echo "Already installed user plugin [{{ name }}]."
        fi
    {% endif %}
{% endfor %}

{% if esMajorVersion < 5 %} {# For the older versions, user plugins are renamed to configured name during install #}
declare CURRENTLY_INSTALLED="{{ allPluginsInstalled }}"
{% else %} {# Custom plugins are not renamed as part of install since 5.0 so we need to record the actual name and not what was configured in plan. #}

# When listing the installed plugins, we remove any plugins that start with a tab character as these plugins are
# part of a plugin bundle. When purging plugins, we would remove the bundle as Elasticsearch will then remove the
# associated bundle plugins
declare CURRENTLY_INSTALLED=$(${ELASTICSEARCH_HOME}/bin/elasticsearch-plugin list | sed '/^\t/d' | sort)
{% endif %}

# This block ensures that if a plugin was removed at the same time as the docker image was upgraded, we
# don't leave config lingering. Mostly to ensure it doesn't create trouble for a future reinstall of that plugin.
# Unlike other cleanups triggered by upgrade, this is done post plugin install in order to compare against real plugin names
if [ "$PURGE_CONFIG_POST_INSTALL" == 'true' ] && [ -e /app/plugins_installed ]; then
    export NEEDS_CONFIG_PURGE=$(comm -23 /app/plugins_installed <(echo "$CURRENTLY_INSTALLED"))
    for PLUGIN in $NEEDS_CONFIG_PURGE; do \
        if [ -e /app/config/$PLUGIN ]; then
            {% if supportsPurge %}
                echo "Detected lingering config for previously installed plugin: $PLUGIN, purging config."
                ${ELASTICSEARCH_HOME}/bin/elasticsearch-plugin remove --purge $PLUGIN
            {% else %}
                echo "Detected lingering config for previously installed plugin, removing: /app/config/$PLUGIN"
                rm -r "/app/config/$PLUGIN"
            {% endif %}
        fi
    done
fi

#Record plugins installed, to enable purging config of any plugin implicitly removed during an in-place upgrade
echo "$CURRENTLY_INSTALLED" > /app/plugins_installed

# Indicate that plugins should be ok for this version, so we don't wipe everything above
touch $PLUGIN_SET_MARKER

echo "Installing user bundles."

{% if not container.shouldInstallUserBundles() %}
echo "Skipping installation of user bundles due to allocation configuration."
{% elif not instance_data.elasticsearch.user_bundles %}
echo "No user bundles defined"
{% else %}
{% for user_bundle in instance_data.elasticsearch.user_bundles %}
    {% set name = user_bundle.name | replace("'", "") %}
    if [ ! -e $USER_BUNDLES_DIR/.found_installed_bundle_{{ name }} ]; then
        echo "Installing bundle [{{ name }}]..."
        {% set clean_bundle_url = user_bundle['url'] | replace("'", "") %}
        {% if clean_bundle_url != user_bundle['url'] %}
          echo "A single quote was filtered from the URL of the bundle. This might result on a container not starting, but could be a signal of an attack"
        {% endif %}
        {% set clean_bundle_url_hash = clean_bundle_url | md5 %}

        if [ ! -e $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip ]; then
          echo "Downloading bundle [{{ name }}]"
          if ! wget --no-use-server-timestamps -O $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip '{{ clean_bundle_url }}'; then
            echo "Could not download bundle [{{ name }}]"
            rm -f $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip
            exit 1
          fi

          echo "Validating bundle [{{ name }}]"
          if ! unzip -t $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip; then
            echo "Bundle [{{ name }}] is not a zip file"
            rm -f $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip
            exit 1
          fi

          echo "Clearing expired cached versions of bundle [{{ name }}]"
          ls -t $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-*.zip | tail -n +3 | xargs rm -f
        fi

        echo "Copying data from bundle [{{ name }}]"
        unzip -o $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip -d $USER_BUNDLES_DIR/

        echo "Installed bundle [{{ name }}]."
        echo "{{ clean_bundle_url_hash }}" > $USER_BUNDLES_DIR/.found_installed_bundle_{{ name }}
    else
        echo "Already installed bundle [{{ name }}]."
    fi
{% endfor %}

echo "Ensuring read/write permissions on all bundles"
chmod u+rw -R $USER_BUNDLES_DIR/*

{% if installBundlesFromStartScript|default(true) %}
# Copy the scripts directory to the found_user_bundles plugin
[[ -h $ES_PATH_CONF/scripts ]] && rm $ES_PATH_CONF/scripts
if [ -e $USER_BUNDLES_SCRIPTS_DIR ]; then
    cp -r $USER_BUNDLES_SCRIPTS_DIR $ES_PATH_CONF/scripts
fi

# Copy the dictionaries directory from the found_user_bundles
[[ -h $ES_PATH_CONF/dictionaries ]] && rm $ES_PATH_CONF/dictionaries
if [ -e $USER_BUNDLES_DICTS_DIR ]; then
    # We copy the dictionaries directly to /app/config. We need `-n` in order not to override existing files
    echo "Installing dictionaries"
    cp -rn $USER_BUNDLES_DICTS_DIR/* $ES_PATH_CONF
fi

{% set store_pass = jvmTrustStorePassword|default("changeit") %}
{% set store_name = jvmTrustStoreName|default("cacerts") %}
SAFE_STORE_NAME='{{ store_name }}'
SAFE_STORE_PASS='{{ store_pass }}'
if [ -e "$USER_BUNDLES_DIR/$SAFE_STORE_NAME" ]; then
    echo "Installing jvm trust store $SAFE_STORE_NAME"
    cp "$USER_BUNDLES_DIR/$SAFE_STORE_NAME" $ES_PATH_CONF
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -Djavax.net.ssl.trustStore=$ES_PATH_CONF/$SAFE_STORE_NAME -Djavax.net.ssl.trustStorePassword=$SAFE_STORE_PASS"
fi

# Maybe an old-style bundle
old_style_items=$(find $USER_BUNDLES_DIR -maxdepth 1 \
        -not -path $USER_BUNDLES_DIR \
        -not -path $USER_BUNDLES_SCRIPTS_DIR \
        -not -path $USER_BUNDLES_DICTS_DIR \
        -not -path "$USER_BUNDLES_DIR/$SAFE_STORE_NAME" \
        -not -path "$USER_BUNDLES_DIR/\.found*")
for item in $old_style_items; do
    echo "Installing old-style-bundle dictionary: $item"
    cp -rn "$item" $ES_PATH_CONF
done

if [ -e $ES_PATH_CONF/krb5.conf ] ; then
   echo "Setting custom kerberos config"
   export ES_JAVA_OPTS="$ES_JAVA_OPTS -Djava.security.krb5.conf=$ES_PATH_CONF/krb5.conf"
fi
{% endif %}
{% endif %}

{% if esMajorVersion >= 5 %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dlog4j2.formatMsgNoLookups=true"
{% endif %}

echo "Done installing plugins and bundles, verifying required config files."

{% if esMajorVersion >= 5 %}
if [ ! -f $XPACK_ROOT/system_key ]; then
    echo "Security System Key file not found ($XPACK_ROOT/system_key)"
    exit 1
fi
{% endif %}

echo "Done verifying required config files."

{% if esMajorVersion == 1 %}
export CONFIG_OPTIONS="-Des.config=/app/config/elasticsearch.yml"
{% elif esMajorVersion == 2 %}
export CONFIG_OPTIONS="-Des.default.path.logs=${LOG_DIR} -Des.default.path.data=${DATA_DIR} -Des.default.path.conf=${ES_PATH_CONF}"
{% elif esMajorVersion == 5 %}
export CONFIG_OPTIONS="-Epath.logs=${LOG_DIR} -Epath.data=${DATA_DIR} -Epath.conf=${ES_PATH_CONF}"
{% else %}
# ES 6
export CONFIG_OPTIONS="-Epath.logs=${LOG_DIR} -Epath.data=${DATA_DIR}"
{% endif %}

# used by metricbeat
export ECE_COMPONENT="elasticsearch"
export ECE_CLUSTER_NAME="{{ container.id().group() }}"
export ECE_INSTANCE_ID="{{ container.id().name() }}"
export ECE_CLUSTER_VERSION="{{ container.versionString() }}"



# set cluster data environment variables, this is done to allow being able to
# modify the elasticsearch environment (for example java configuration and the like)
# without having to vacate the node, as Docker environment properties can't be modified
# while these properties are set by the allocator
{% for environment_variable in cluster_data.overrides.elasticsearchEnv %}
export {{environment_variable}}
{% endfor %}

echo "$(date +%Y-%m-%dT%H:%M:%S%z)" >> {{ script_ready_lock }}
echo "Script ready lock {{ script_ready_lock }} created"

# Allocator creates this file when there are tasks that need to be completed
# before we can start elasticsearch (e.g., keystore update).
# If file exists, wait (up to five minutes) for it to go away.
{% set lock_file = startupLock|default("_no_lock_file") %}
{% set lock_timeout = startupLockTimeout|default(300) %}
waited=0
while [ -e {{ lock_file }} ] ; do
  if [ $waited -ge {{ lock_timeout }} ] ; then
    echo "GIVING UP on startup lock after {{ lock_timeout }} seconds."
    break
  fi
  if [ $((waited%10)) = 0 ] ; then
    echo "Waiting for startup lock ($waited seconds elapsed)."
  fi
  sleep 1
  ((++waited))
done

run_beat () {
  BEAT_NAME=${1}
  # Remove any lockfile left over from a container restart
  rm -f /tmp/${BEAT_NAME}-data/${BEAT_NAME}.lock || true
  while true; do
    /opt/${BEAT_NAME}/${BEAT_NAME} --path.config /app/config/${BEAT_NAME} --path.data /tmp/${BEAT_NAME}-data
    echo "${BEAT_NAME} exited with code $?. Re-running in 2 seconds"
    sleep 2
  done
}

{% if instance_data.observability.logging is defined %}
echo "Starting filebeat"
run_beat filebeat &
{% endif %}
{% if instance_data.observability.metrics is defined %}
echo "Starting metricbeat"
run_beat metricbeat &
{% endif %}

echo "Starting Elasticsearch after waiting on startup lock {{ lock_file }} for $waited seconds. See Elasticsearch logs for further output."

# restore stdout and stderr
exec 1>&3 2>&4

# Do not abort script if elasticsearch returns error code
set +e

_term() {
   echo "Caught SIGTERM"
   kill -TERM $(cat /app/es.pid)
}

trap _term TERM

{% if esMajorVersion <= 2 %}
/elasticsearch/bin/elasticsearch -Dfound.cluster_name={{ container.id().group() }} -Dfound.instance_name={{ container.id().name() }} -p /app/es.pid $CONFIG_OPTIONS $*
{% else %}
${ELASTICSEARCH_HOME}/bin/elasticsearch -p /app/es.pid $CONFIG_OPTIONS $* &
{% endif %}

ES_PID=$!
wait ${ES_PID}
ES_EXIT_CODE=$?
trap - TERM
wait ${ES_PID}
# The double wait above is deliberate as the first one returns immediately after TERM signal gets trapped.
# Ref: https://github.com/elastic/cloud-assets/pull/451
if [ "$ES_EXIT_CODE" -ge "1" ]; then
    echo $(date +%Y-%m-%dT%H:%M:%S%z) "Elasticsearch exited with status $ES_EXIT_CODE, running ./on-error-exitcode.sh $ES_EXIT_CODE" | tee -a $BOOT_LOG_FILE
    ./on-error-exitcode.sh $ES_EXIT_CODE
fi

# Remove the script ready lock
rm {{ script_ready_lock }}

exit $ES_EXIT_CODE
